from torch import nn
from torch.nn import functional as F
from transformers.modeling_outputs import SequenceClassifierOutput
from ankh.models import layers


class ConvBertForBinaryClassification(layers.BaseModule):
    def __init__(
        self,
        input_dim: int,
        nhead: int,
        hidden_dim: int,
        num_hidden_layers: int = 1,
        num_layers: int = 1,
        kernel_size: int = 7,
        dropout: float = 0.2,
        pooling: str = "max",
    ):
        if pooling is None:
            raise ValueError(
                '`pooling` cannot be `None` in a binary classification task. Expected ["avg", "max"].'
            )

        super(ConvBertForBinaryClassification, self).__init__(
            input_dim=input_dim,
            nhead=nhead,
            hidden_dim=hidden_dim,
            num_hidden_layers=num_hidden_layers,
            num_layers=num_layers,
            kernel_size=kernel_size,
            dropout=dropout,
            pooling=pooling,
        )

        """
            ConvBert model for binary classification task.

            Args:
                input_dim: Dimension of the input embeddings.
                nhead: Integer specifying the number of heads for the `ConvBert` model.
                hidden_dim: Integer specifying the hidden dimension for the `ConvBert` model.
                num_hidden_layers: Integer specifying the number of hidden layers for the `ConvBert` model.
                num_layers: Integer specifying the number of `ConvBert` layers.
                kernel_size: Integer specifying the filter size for the `ConvBert` model. Default: 7
                dropout: Float specifying the dropout rate for the `ConvBert` model. Default: 0.2
                pooling: String specifying the global pooling function. Accepts "avg" or "max". Default: "max"
        """

        self.decoder = nn.Linear(input_dim, 1)
        self.init_weights()

    def init_weights(self):
        initrange = 0.1
        self.decoder.bias.data.zero_()
        self.decoder.weight.data.uniform_(-initrange, initrange)

    def _compute_loss(self, logits, labels):
        if labels is not None:
            loss = F.binary_cross_entropy_with_logits(logits, labels)
        else:
            loss = None
        return loss

    def forward(self, embed, labels=None):
        hidden_inputs = self.convbert_forward(embed)
        hidden_inputs = self.pooling(hidden_inputs)
        logits = self.decoder(hidden_inputs)
        loss = self._compute_loss(logits, labels)

        return SequenceClassifierOutput(
            loss=loss, logits=logits, hidden_states=None, attentions=None
        )



import argparse
import pathlib
from ankh.models import ankh_transformers
from ankh import utils
import torch
from tqdm.auto import tqdm
import numpy as np


def create_parser() -> argparse.ArgumentParser:
    argparser = argparse.ArgumentParser()
    argparser.add_argument(
        "--model",
        type=str,
        help="Whether to use the base model or the large model.",
    )
    argparser.add_argument(
        "--fasta_path", type=str, help="Location to the fasta file."
    )
    argparser.add_argument(
        "--output_path", type=str, help="Location to save the embeddings."
    )
    argparser.add_argument(
        "--use_gpu", type=bool, help="Whether to use GPU or not."
    )
    return argparser


def validate_output_path(path: pathlib.Path) -> None:
    if not path.exists():
        raise FileNotFoundError(f"File not found. Recieved path: {path}")


def get_device(use_gpu):
    if torch.cuda.is_available() and use_gpu:
        device = torch.device("cuda:0")
    else:
        device = torch.device("cpu")

    return device


def main(args: argparse.Namespace) -> None:
    model, tokenizer = ankh_transformers.load_model(args.model)
    fasta_dataset = utils.FastaDataset(args.fasta_path)
    output_path = pathlib.Path(args.output_path)
    validate_output_path(output_path)

    device = get_device(args.use_gpu)

    shift_left = 0
    shift_right = -1

    model.eval()
    model.to(device=device)

    with torch.no_grad():
        for idx, sample in enumerate(tqdm(fasta_dataset), start=1):
            ids = tokenizer.batch_encode_plus(
                [sample],
                add_special_tokens=True,
                padding=True,
                is_split_into_words=True,
                return_tensors="pt",
            )
            embedding = model(input_ids=ids["input_ids"].to(device))[0]
            current_embeddings = (
                embedding[0].cpu().numpy()[shift_left:shift_right]
            )
            np.save(output_path / f"sequence_{idx}", current_embeddings)


if __name__ == "__main__":
    parser = create_parser()
    args = parser.parse_args()
    main(args)



from torch import nn
from functools import partial
import torch
import transformers.models.convbert as c_bert


class GlobalMaxPooling1D(nn.Module):
    def __init__(self):
        """
            Applies global max pooling over timesteps dimension
        """

        super().__init__()
        self.global_max_pool1d = partial(torch.max, dim=1)

    def forward(self, x):
        out, _ = self.global_max_pool1d(x)
        return out


class GlobalAvgPooling1D(nn.Module):
    def __init__(self):
        """
            Applies global average pooling over timesteps dimension
        """

        super().__init__()
        self.global_avg_pool1d = partial(torch.mean, dim=1)

    def forward(self, x):
        out = self.global_avg_pool1d(x)
        return out


class BaseModule(nn.Module):
    def __init__(
        self,
        input_dim: int,
        nhead: int,
        hidden_dim: int,
        num_hidden_layers: int = 1,
        num_layers: int = 1,
        kernel_size: int = 7,
        dropout: float = 0.2,
        pooling: str = None,
    ):
        """
        Base model that consists of ConvBert layer.

        Args:
            input_dim: Dimension of the input embeddings.
            nhead: Integer specifying the number of heads for the `ConvBert` model.
            hidden_dim: Integer specifying the hidden dimension for the `ConvBert` model.
            nlayers: Integer specifying the number of layers for the `ConvBert` model.
            kernel_size: Integer specifying the filter size for the `ConvBert` model. Default: 7
            dropout: Float specifying the dropout rate for the `ConvBert` model. Default: 0.2
            pooling: String specifying the global pooling function. Accepts "avg" or "max". Default: "max"
        """
        super().__init__()

        self.model_type = "Transformer"
        encoder_layers_Config = c_bert.ConvBertConfig(
            hidden_size=input_dim,
            num_attention_heads=nhead,
            intermediate_size=hidden_dim,
            conv_kernel_size=kernel_size,
            num_hidden_layers=num_hidden_layers,
            hidden_dropout_prob=dropout,
        )

        self.transformer_encoder = nn.ModuleList(
            [c_bert.ConvBertLayer(encoder_layers_Config) for _ in range(num_layers)]
        )

        if pooling is not None:
            if pooling in {"avg", "mean"}:
                self.pooling = GlobalAvgPooling1D()
            elif pooling == "max":
                self.pooling = GlobalMaxPooling1D()
            else:
                raise ValueError(
                    f"Expected pooling to be [`avg`, `max`]. Recieved: {pooling}"
                )

    def convbert_forward(self, x):
        for convbert_layer in self.transformer_encoder:
            x = convbert_layer(x)[0]
        return x



from torch import nn
from torch.nn import functional as F
from transformers.modeling_outputs import TokenClassifierOutput
from ankh.models import layers


class ConvBertForMultiLabelClassification(layers.BaseModule):
    def __init__(
        self,
        num_tokens: int,
        input_dim: int,
        nhead: int,
        hidden_dim: int,
        num_hidden_layers: int = 1,
        num_layers: int = 1,
        kernel_size: int = 7,
        dropout: float = 0.2,
    ):
        super(ConvBertForMultiLabelClassification, self).__init__(
            input_dim=input_dim,
            nhead=nhead,
            hidden_dim=hidden_dim,
            num_hidden_layers=num_hidden_layers,
            num_layers=num_layers,
            kernel_size=kernel_size,
            dropout=dropout,
            pooling=None,
        )
        """
            ConvBert model for multilabel classification task.
            Args:
                num_tokens: Integer specifying the number of tokens that should be the output of the final layer.
                input_dim: Dimension of the input embeddings.
                nhead: Integer specifying the number of heads for the `ConvBert` model.
                hidden_dim: Integer specifying the hidden dimension for the `ConvBert` model.
                num_hidden_layers: Integer specifying the number of hidden layers for the `ConvBert` model.
                num_layers: Integer specifying the number of `ConvBert` layers.
                kernel_size: Integer specifying the filter size for the `ConvBert` model. Default: 7
                dropout: Float specifying the dropout rate for the `ConvBert` model. Default: 0.2
        """

        self.model_type = "Transformer"
        self.num_labels = num_tokens
        self.decoder = nn.Linear(input_dim, num_tokens)
        self.init_weights()

    def init_weights(self):
        initrange = 0.1
        self.decoder.bias.data.zero_()
        self.decoder.weight.data.uniform_(-initrange, initrange)

    def _compute_loss(self, logits, labels):
        if labels is not None:
            loss = F.binary_cross_entropy_with_logits(
                logits.view(-1, self.num_labels), labels.view(-1, self.num_labels)
            )
        else:
            loss = None
        return loss

    def forward(self, embed, labels=None):
        hidden_inputs = self.convbert_forward(embed)
        logits = self.decoder(hidden_inputs)
        loss = self._compute_loss(logits, labels)

        return TokenClassifierOutput(
            loss=loss,
            logits=logits,
            hidden_states=None,
            attentions=None,
        )



from transformers import (
    T5EncoderModel,
    T5ForConditionalGeneration,
    AutoTokenizer,
    TFT5EncoderModel,
    TFT5ForConditionalGeneration,
)
from enum import Enum
from typing import List, Tuple, Union


class AvailableModels(Enum):
    """Ankh pre-trained model paths."""

    ANKH_BASE = "ElnaggarLab/ankh-base"
    ANKH_LARGE = "ElnaggarLab/ankh-large"


def get_available_models() -> List:
    """Returns a `list` of the current available pretrained models.

    Returns:
        List: Available models.
    """
    return [o.name.lower() for o in AvailableModels]


def load_tf_model(path, generation, output_attentions):
    if generation:
        return TFT5ForConditionalGeneration.from_pretrained(
            path, output_attentions=output_attentions, from_pt=True
        )
    else:
        return TFT5EncoderModel.from_pretrained(
            path, output_attentions=output_attentions, from_pt=True
        )


def load_pt_model(path, generation, output_attentions):
    if generation:
        return T5ForConditionalGeneration.from_pretrained(
            path, output_attentions=output_attentions
        )
    else:
        return T5EncoderModel.from_pretrained(
            path, output_attentions=output_attentions
        )


def load_onnx_model(path, generation, output_attentions):
    raise NotImplementedError('`load_onnx_model` is not implemented yet.')


class SupportedModelFormatFunctions(Enum):
    PT = load_pt_model
    TF = load_tf_model
    ONNX = load_onnx_model


def get_supported_model_format_functions() -> List:
    """Returns a `list` of the current supported model formats.

    Returns:
        List: Supported model formats.
    """
    return [o.name.lower() for o in SupportedModelFormatFunctions]


def get_specified_model(path: str,
                        generation: bool,
                        output_attentions: bool,
                        model_format: str):

    model_loading_function = getattr(SupportedModelFormatFunctions,
                                     model_format.upper())

    return model_loading_function(path,
                                  generation=generation,
                                  output_attentions=output_attentions)


def load_base_model(
    generation: bool = False,
    output_attentions: bool = False,
    model_format: str = 'pt',
) -> Tuple[T5EncoderModel, AutoTokenizer]:
    """Downloads and returns the base model and its tokenizer

    Args:
        generation (bool, optional): Whether to return
                                     `T5ForConditionalGeneration` will be
                                     returned otherwise `T5EncoderModel` will
                                     be returned. Defaults to False.
        output_attentions (bool, optional): Whether to return the attention or
                                            not. Defaults to False.

        model_format (str, optional): The model format, currently supports
                                      'tf' and 'pt'. Defaults to 'pt'.

    Returns:
        Tuple[Union[T5EncoderModel, T5ForConditionalGeneration],
        AutoTokenizer]: Returns T5 Model and its tokenizer.
    """
    tokenizer = AutoTokenizer.from_pretrained(AvailableModels.ANKH_BASE.value)
    model = get_specified_model(
        path=AvailableModels.ANKH_BASE.value,
        generation=generation,
        output_attentions=output_attentions,
        model_format=model_format,
    )
    return model, tokenizer


def load_large_model(
    generation: bool = False,
    output_attentions: bool = False,
    model_format='pt',
) -> Tuple[Union[T5EncoderModel, T5ForConditionalGeneration], AutoTokenizer]:
    """Downloads and returns the large model and its tokenizer

    Args:
        generation (bool, optional): Whether to return
                                     `T5ForConditionalGeneration` will be
                                     returned otherwise `T5EncoderModel` will
                                     be returned. Defaults to False.
        output_attentions (bool, optional): Whether to return the attention or
                                            not. Defaults to False.
        model_format (str, optional): The model format, currently supports
                                      'tf' and 'pt'. Defaults to 'pt'.

    Returns:
        Tuple[Union[T5EncoderModel, T5ForConditionalGeneration],
        AutoTokenizer]: Returns T5 Model and its tokenizer.
    """
    tokenizer = AutoTokenizer.from_pretrained(AvailableModels.ANKH_LARGE.value)
    model = get_specified_model(
        path=AvailableModels.ANKH_LARGE.value,
        generation=generation,
        output_attentions=output_attentions,
        model_format=model_format,
    )
    return model, tokenizer


available_models_fns = {"base": load_base_model, "large": load_large_model}


def load_model(
    model_name: str, generation: bool = False, output_attentions: bool = False
) -> Tuple[Union[T5EncoderModel, T5ForConditionalGeneration], AutoTokenizer]:
    """Downloads and returns the specified model and its tokenizer
    Args:
        model_name (str): Model name, Expects "base" or "large"
        generation (bool, optional): Whether to return
                                     `T5ForConditionalGeneration` will be
                                     returned otherwise `T5EncoderModel` will
                                     be returned. Defaults to False.
        output_attentions (bool, optional): Whether to return the attention or
                                            not. Defaults to False.
    Returns:
        Tuple[Union[T5EncoderModel, T5ForConditionalGeneration],
        AutoTokenizer]: Returns T5 Model and its tokenizer.
    """

    return available_models_fns[model_name](
        generation=generation, output_attentions=output_attentions
    )



from ankh.models.convbert_binary_classification import (
    ConvBertForBinaryClassification,
)
from ankh.models.convbert_multiclass_classification import (
    ConvBertForMultiClassClassification,
)
from ankh.models.convbert_multilabel_classification import (
    ConvBertForMultiLabelClassification,
)
from ankh.models.convbert_regression import ConvBertForRegression

from .ankh_transformers import (
    get_available_models,
    load_base_model,
    load_large_model,
    load_model,
)



from .models import get_available_models, load_base_model, load_large_model, load_model

from .utils import FastaDataset, CSVDataset

from .models import (
    ConvBertForBinaryClassification,
)
from .models import (
    ConvBertForMultiClassClassification,
)
from .models import ConvBertForRegression
from .models import ConvBertForMultiLabelClassification
from typing import Union


available_tasks = {
    "binary": ConvBertForBinaryClassification,
    "regression": ConvBertForRegression,
    "multiclass": ConvBertForMultiClassClassification,
    "multilabel": ConvBertForMultiLabelClassification,
}


def get_available_tasks():
    return list(available_tasks.keys())


def load_downstream_model(
    task,
) -> Union[
    ConvBertForBinaryClassification,
    ConvBertForMultiClassClassification,
    ConvBertForRegression,
    ConvBertForMultiLabelClassification
]:
    return available_tasks[task]


__version__ = "1.0"



from torch.utils.data import Dataset
import pathlib

try:
    from Bio import SeqIO as seqio
except ImportError:
    seqio = None

import torch
import numpy as np
import random


def set_seeds(seed: int = 7):
    torch.manual_seed(seed)
    np.random.seed(seed)
    random.seed(seed)


class FastaDataset(Dataset):
    def __init__(self, fasta_path: str):
        """
        A class for loading a FASTA file.

        Args:
            fasta_path: String specifying the path of the FASTA file
        """
        self.fasta_path = pathlib.Path(fasta_path)
        if not self.fasta_path.exists():
            raise FileNotFoundError(
                f"Fasta file does not exist. Recieved path: {fasta_path}."
            )

        if seqio is None:
            raise ImportError(
                "In order to use `FastaDataset` class you need `biopython` installed. "
                "Install it using `pip install biopython`."
            )

        self.sequences = [
            list(line.seq) for line in seqio.parse(self.fasta_path, "fasta")
        ]

    def __len__(self):
        return len(self.sequences)

    def __getitem__(self, index):
        return self.sequences[index]


class CSVDataset(Dataset):
    def __init__(self, dataframe, sequences_column_name, labels_column_name):
        """
        A class for loading a CSV file.

        Args:
            dataframe: `Dataself.Frame` instance that contains the sequences and labels.
            sequences_column_name: String specifying the column name of the sequences.
            labels_column_name: String specifying the column name of the labels.
        """
        self.df = dataframe
        self.seq_col_name = sequences_column_name
        self.label_col_name = labels_column_name

    def __len__(self):
        return self.df.shape[0]

    def __getitem__(self, index):
        return (
            list(self.df.iloc[index, self.seq_col_name]),
            self.df.iloc[index, self.label_col_name],
        )
